%% Adaptive coder
clc
clear all

% Read avi file
% Available movies are 
% 'screen_content_1.avi', 352x288 resolution, 73 frames
% 'screen_content_2.avi', 720x480 resolution, 129 frames
% 'foreman_cif.avi', 352x288 resolution, 300 frames
movObj = mmreader('screen_content_1.avi'); 

% Convert to 4D matrix with coefficients
vidFrames = read(movObj);

display('Adaptive coding')

% Find height, width and number of frames
sizeOfMatrix = size(vidFrames);
nHeight = sizeOfMatrix(1);
nWidth = sizeOfMatrix(2);
nFrames = sizeOfMatrix(4);

% Set I-frame frequency (GOP size) and lambda (Lagrange 
% multiplier) in addition to the quantization parameter 
% (totalq) and amount of bits sent to the extended 
% end-zone quantizer.  
totalq = 32;
quant_factor = 4;
iFrameFreq = 6;
lambda = 0.85;

% Size of blocks to be coded
bSize = 4;
      
% Declare variables that can contain information about frames
rFr = uint8(zeros(nHeight, nWidth,nFrames));
gFr = uint8(zeros(nHeight, nWidth,nFrames));
bFr = uint8(zeros(nHeight, nWidth,nFrames));

yFrames = zeros(nHeight, nWidth,nFrames);
uFr = uint8(zeros(nHeight, nWidth,nFrames));
vFr = uint8(zeros(nHeight, nWidth,nFrames));

yCells = cell(1,nFrames);
rsqYCells = cell(1,nFrames);
tradYClls = cell(1,nFrames);
uCells = cell(1,nFrames);
vCells = cell(1,nFrames);

% Split video into R, G and B frames
for i = 1:nFrames
  rFr(:,:,i) = vidFrames(:,:,1,i);
  gFr(:,:,i) = vidFrames(:,:,2,i);
  bFr(:,:,i) = vidFrames(:,:,3,i);
  tmpR = rFr(:,:,i);
  tmpG = gFr(:,:,i);
  tmpB = bFr(:,:,i);
  % Convert from RGB to YUV
  [yFrames(:,:,i),uFr(:,:,i),vFr(:,:,i)]=rgb2yuv(tmpR,tmpG,tmpB);    
end

% Declare variables to pre-allocate memory
blocks = zeros(bSize,bSize,nWidth*nHeight/(bSize.^2));
rcTrdYFrs = zeros(size(yFrames)); 
rcRSQYFrs = zeros(size(yFrames));
rcYFrs = zeros(size(yFrames));
dpcmBlocks = zeros(size(blocks));
quantTradBlocks = zeros(size(blocks));
iQuantTradBlocks = zeros(size(blocks));
rcTrdBlcks = zeros(size(blocks));
tempBlock = zeros(bSize,bSize);
tempBlockTrad = zeros(bSize,bSize);
rdTrad = zeros(2,length(blocks));
rdRSQ = zeros(2,length(blocks));
rcTmpFrm = zeros(bSize,bSize);

% These will contain the PSNR values of the reconstructed frames
%for each coder.
psnrVector = zeros(1,nFrames);
rsqPSNRVector = zeros(1,nFrames);
tradPSNRVector = zeros(1,nFrames);

% These  variables contain the total bits used for each coder.
totalRateRSQ = 0;
totalrateTrad = 0;
totalRateAdaptive = 0;

% Iterate over all frames and encode each frame 
for i = 1:nFrames    
  l = 1;
  j = bSize;
  k = bSize;
  % Split frames into blocks and create a 3D matrix
  % to contain all blocks.
  for p = bSize:bSize:nWidth*nHeight/bSize                    
    blocks(:,:,l) = yFrames(k-(bSize-1):k,j-(bSize-1):j,i);
    k = k + bSize;
      if mod(p,nHeight) == 0
        j = j+bSize;
        k = bSize;
      end
      l = l+1;
  end        
  % Differential coding. Skip differential coding for initial
  % frame and each I-frame, determined by I-frame frequency 
  % (iFrameFreq).
  if i == 1 || mod(i,iFrameFreq) == 0                
    dpcmBlocks = blocks;        
    anchorBlocks = blocks;        
  else                
    dpcmBlocks = blocks - anchorBlocks;                
  end
    
  rateTrad = 0;
  % Traditional coder, iterate over each block in the frame.
  for p = 1:length(dpcmBlocks)
    % 2D DCT transform and uniform quantization done in
    % quantizer(blocks, qp) function.
    quantTradBlocks(:,:,p) = quantizer(dpcmBlocks(:,:,p),totalq);        
        
    % Entropy code quantized blocks
    % Extract minimum value and add it to the values to be 
    % entropy coded, to ensure that no values are less than 0. 
    huffBlock = quantTradBlocks(:,:,p);
    minVal = abs(min(huffBlock(:)));
    [tradCell,info]=norm2huff(uint8(huffBlock(:)+minVal));
    % We split the huffcodes into indices and huff values, the
    % indices will not be used.
    [tmpk,tmpl,huff] = find(info.huffcodes);
    % We pass the huffcodes to huffcodes2bin to create a list 
    % of bit codes.
    [listOfBitCodes,symbols] = huffcodes2bin(info.huffcodes);        
    sortedCodes = zeros(1,length(huff));
    % Sort frequency of codes to match huffman bit codes
    % (most frequent => shortest bit codes)
    for j = 1:length(huff)
      sortedCodes(j) = sum(sum(quantTradBlocks(:,:,p)==huff(j)));
    end        
    sortedCodes = sort(sortedCodes,'descend');
    bitsPrBlck = 0;     
    tmpBits = 0;
    % Multiply the length of bit codes with the frequency of the 
    % value to get the amount of bits used. Add the 
    % bit-representation of the actual value.
    for j = 1:length(huff)
      tmpBits = length(listOfBitCodes{j})*sortedCodes(j);
      bitsPrBlck = bitsPrBlck+tmpBits+ceil(log2(huff(j)));
    end            
    if sum(tradCell(:)) == 0 
      % If the sum of rsqCell is 0, all values in the Huffman
      % coded block are equal and the Huffman coder does not 
      % work properly, consequently we have to add the value 
      % (minVal) of the block that was coded to the inversely
      % entropy coded block.                
      rcTmpFrm(:)=minVal;  
      tempTradBlock = quantTradBlocks(:,:,p);
      % If the uniform value is non-zero, add the given 
      % bit-representation of the value to the amount
      % of bits used. Otherwise add 1 bit.
      if var(tempTradBlock(:)) ~= 0
        bitsPrBlck=bitsPrBlck+ceil(log2(minVal));
      else
        bitsPrBlck = bitsPrBlck + 1;
      end
    else
      % Inverse entropy coding and add the minVal which was 
      % subtracted before the entropy coding was done.
      rcTmpFrm(:) = double(huff2norm(tradCell,info))-minVal;         
    end                
    rateTrad = rateTrad + bitsPrBlck;
    % Inverse quantization
    iQuantTradBlocks(:,:,p) = iQuantizer(rcTmpFrm,totalq); 
    % Store the rate for each block
    rdTrad(1,p) = bitsPrBlck;
  end      
   
  rateRSQ = 0;
  quantizedRSQBlocks = zeros(size(blocks));
  iQuantRSQBlocks = zeros(size(blocks));
  % Perform quantization on entire frame unless it consist of 
  % all 0 coefficients, which will cause the coder to break down.
  if var(dpcmBlocks(:) ~= 0)
    tempVector = quantize_rsq(dpcmBlocks(:), quant_factor);   
    quantizedRSQBlocks(:) = tempVector(:);
  else
    quantizedRSQBlocks(:) = 0;
  end
  
  % RSQ coding
  for p = 1:length(quantizedRSQBlocks)      
    % Split block into plus and minus values, which are coded
    % seperately and independently.
    plusIndices = find(quantizedRSQBlocks(:,:,p)>=0);
    minIndices = find(quantizedRSQBlocks(:,:,p)<0);
    tempFrame = quantizedRSQBlocks(:,:,p);
    minVal = abs(min(tempFrame(:)));     
    plusValues = tempFrame(plusIndices);
    minVals = tempFrame(minIndices); 
    bitsPrBlck = 0;  
    % Entropy code positive quantized values (if any)
    if ~isempty(plusValues) 
      [rsqCell, info] = norm2huff(uint8(plusValues(:)));
      [tmpk,tmpl,huff] = find(info.huffcodes);
      [listOfBitCodes,symbols] = huffcodes2bin(info.huffcodes);
      sortedCodes = zeros(1,length(huff));        
      % Sort frequency of codes to match huffman bit codes 
      % (most frequent => shortest bit codes)
      for j = 1:length(huff)
        sortedCodes(j) = sum(sum(plusValues(:,:)==huff(j)));
      end        
      sortedCodes = sort(sortedCodes,'descend');
      tmpBits = 0;
      % Multiply the length of bit codes with the frequency of 
      % the value to get the amount of bits used. Add the 
      % bit-representation of the actual value.
      for j = 1:length(huff)    
        tmpBits = length(listOfBitCodes{j})*sortedCodes(j);
        bitsPrBlck=bitsPrBlck+tmpBits+ceil(log2(huff(j)));
      end
    end
    % Entropy code negative quantized values (if any)
    if ~isempty(minVals)                     
      tempMinus = minVals + minVal;
      [minusCell, minusInfo] = norm2huff(uint8(tempMinus(:)));        
      [k2,l2,minHuff] = find(minusInfo.huffcodes);    
      minListOfBitCodes = huffcodes2bin(minusInfo.huffcodes);      
      minSortCodes = zeros(1,length(minHuff));
      % Sort frequency of codes to match huffman bit codes 
      %(most frequent => shortest bit codes)
      for j = 1:length(minHuff)
        minSortCodes(j) = sum(sum(tempMinus==minHuff(j)));
      end     
      minSortCodes = sort(minSortCodes,'descend'); 
      tmpBits = 0;
      % Multiply the length of bit codes with the frequency of 
      % the value to get the amount of bits used. Add the 
      % bit-representation of the actual value.
      for j = 1:length(minHuff)        
        tmpBits=length(minListOfBitCodes{j})*minSortCodes(j);
        bitsPrBlck=bitsPrBlck+tmpBits+ceil(log2(minHuff(j)));
      end                  
    end              
    tempBlock = zeros(bSize,bSize);
    rcTmpFrm = zeros(bSize,bSize); 
    % Inversely entropy code eventual positive values.
    if ~isempty(plusIndices)
      % If the sum of rsqCell is 0, all values in the Huffman 
      % coded block are equal and the Huffman coder does not
      % work properly, consequently we have to add the value 
      % (minVal) of the block that was coded.                   
      if sum(rsqCell(:)) == 0 
        % Entire block consists of minVal  
        tempBlock(:) = double(minVal); 
        % Add one bit for each 0-block for comparison.
        bitsPrBlck =  bitsPrBlck + 1;
      else
        % Create a block (tmp) which holds the inversely 
        % entropy coded values.
        tmp = double(huff2norm(uint8(rsqCell(:)),info));
        tempBlock(1:length(tmp)) = double(tmp);          
      end
      % Insert the positive values in their correct positions.
      rcTmpFrm(plusIndices)=tempBlock(1:length(plusIndices));
    end                                          
    % Inversely entropy code eventual negative values.
    if ~isempty(minIndices)      
      minVals = double(huff2norm(uint8(minusCell(:)),minusInfo));
      % Insert negative values in their correct positions.
      rcTmpFrm(minIndices)=minVals(1:length(minIndices))-minVal;            
    end                                    
    if bitsPrBlck == 0 % Block is all zeros
      bitsPrBlck = bitsPrBlck + 1;
    end
    rateRSQ = rateRSQ + bitsPrBlck;            
    iQuantRSQBlocks(:,:,p) = rcTmpFrm;        
    rdRSQ(1,p) = bitsPrBlck;        
  end         
    
  % Inverse differential coding
  if i == 1 || mod(i,iFrameFreq) == 0                                 
    rcTrdBlcks = iQuantTradBlocks;        
    recAnchorBlocks = iQuantTradBlocks;
    rcRSQBlcks = iQuantRSQBlocks;
    rsqAnchorBlocks = iQuantRSQBlocks;
  else                 
    rcTrdBlcks = iQuantTradBlocks + recAnchorBlocks; 
    rcRSQBlcks = iQuantRSQBlocks + rsqAnchorBlocks;      
  end    
    
    
  % Reconstruct frame
  j = bSize;
  k = bSize;
  adaptiveRate = 0;
  for p = 1:length(blocks)
    % Calculate distortion for RDO, traditional coder
    dist=sum(sum(rcTrdBlcks(:,:,p)-blocks(:,:,p))).^2;         
    rdTrad(2,p) = dist;
    % Calculate distortion for RDO, RSQ coder
    dist=sum(sum(rcRSQBlcks(:,:,p) - blocks(:,:,p))).^2;         
    rdRSQ(2,p) = dist;
    % Calculate Lagrange values
    lagrangeRSQ = rdRSQ(2,p)+lambda.*rdRSQ(1,p);
    lagrangeTrad = rdTrad(2,p) + lambda.*rdTrad(1,p);
    % Insert the block with the lowest Lagrange value
    if lagrangeRSQ < lagrangeTrad      
      rcYFrs(k-(bSize-1):k,j-(bSize-1):j,i)=rcRSQBlcks(:,:,p);
      adaptiveRate = adaptiveRate + rdRSQ(1,p);
    else        
      rcYFrs(k-(bSize-1):k,j-(bSize-1):j,i)=rcTrdBlcks(:,:,p);
      adaptiveRate = adaptiveRate + rdTrad(1,p);
    end        
    % Create RSQ and traditional coder frames for comparison
    % purposes.
    rcTrdYFrs(k-(bSize-1):k,j-(bSize-1):j,i)=rcTrdBlcks(:,:,p);
    rcRSQYFrs(k-(bSize-1):k,j-(bSize-1):j,i)=rcRSQBlcks(:,:,p);
    k = k + bSize;
    if mod(p,nHeight/bSize) == 0
      j = j+bSize;
      k = bSize;
    end
  end               
  
  % Print PSNR and bits used for each coder
  tradPSNR = psnrcalc(yFrames(:,:,i),rcTrdYFrs(:,:,i));    
  tradPSNRVector(i) = tradPSNR;
  totalrateTrad = totalrateTrad + rateTrad;
  fprintf('Traditional coder PSNR: %d \n',tradPSNR);
  fprintf('Traditional coder rate: %d \n',rateTrad);  
    
  rsqPSNR = psnrcalc(yFrames(:,:,i),rcRSQYFrs(:,:,i));    
  rsqPSNRVector(i) = rsqPSNR;
  totalRateRSQ = totalRateRSQ + rateRSQ;
  fprintf('RSQ coder PSNR: %d \n',rsqPSNR);
  fprintf('RSQ coder rate: %d \n',rateRSQ);
    
  adaptivePSNR = psnrcalc(yFrames(:,:,i),rcYFrs(:,:,i));    
  psnrVector(i) = adaptivePSNR;
  totalRateAdaptive = totalRateAdaptive + adaptiveRate;
  fprintf('Adaptive coder PSNR: %d \n',adaptivePSNR);
  fprintf('Adaptive coder rate: %d \n',adaptiveRate);
  fprintf('New frame\n');    
end

% Create yuv file
for j = 1:nFrames    
  % Convert to cells for yuv exporting
  yCells(j) = mat2cell(rcYFrs(:,:,j),nHeight,nWidth);
  rsqYCells(j) = mat2cell(rcRSQYFrs(:,:,j),nHeight,nWidth);
  tradYClls(j) = mat2cell(rcTrdYFrs(:,:,j),nHeight,nWidth);
  uCells(j) = mat2cell(uFr(:,:,j),nHeight,nWidth);
  vCells(j) = mat2cell(vFr(:,:,j),nHeight,nWidth);
end

% Export to yuv file
yuv_export(yCells,uCells,vCells,'output.yuv',nFrames,'w');
yuv_export(rsqYCells,uCells,vCells,'output_rsq.yuv',nFrames,'w');
yuv_export(tradYClls,uCells,vCells,'output_trd.yuv',nFrames,'w');